package com.vernaio.example;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.IntStream;

import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.fasterxml.jackson.datatype.jdk8.Jdk8Module;

import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoFloatDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoFloatDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.exception.EngineException;
import com.vernaio.aivis.engine.v2.base.sdk.exception.MissingLibraryException;
import com.vernaio.aivis.engine.v2.base.sdk.exception.WrongFlavourException;
import com.vernaio.aivis.engine.v2.signal.prediction.sdk.SignalPredictionData;

public class Utils {
	
	private static final ObjectMapper MAPPER = new ObjectMapper().registerModule(new Jdk8Module());
	
	public static class FloatSignal {
		
		private final String signalId;
		
		private final List<DtoFloatDataPoint> dataPoints;
		
		protected FloatSignal(final String signalId) {
			this.signalId = signalId;
			this.dataPoints = new ArrayList<DtoFloatDataPoint>();
		}
		
		public List<DtoFloatDataPoint> getDataPoints() {
			return this.dataPoints;
		}
		
		public String getSignalId() {
			return signalId;
		}
	}
	
	public static <T> T readJson(final File file, final Class<T> clazz) throws JsonParseException, JsonMappingException, IOException {
		System.out.println(String.format("\t... reading file '%s' ...", file.getPath()));
		return MAPPER.readValue(file, clazz);
	}
	
	public static void writePredictions(final File file, final List<? extends IDtoFloatDataPoint> scores) throws FileNotFoundException, IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		
		final ObjectNode json = MAPPER.createObjectNode();
		json.put("_type", "Float");
		final ArrayNode dataPointsNode = json.putArray("dataPoints");
		for (IDtoFloatDataPoint dataPoint : scores) {
			final ObjectNode dataPointNode = MAPPER.createObjectNode();
			dataPointNode.put("timestamp", dataPoint.getTimestamp());
			dataPointNode.put("value", dataPoint.getValue());
			dataPointsNode.add(dataPointNode);
		}
		
		final FileWriter fileWriter = new FileWriter(file);
		fileWriter.write(json.toPrettyString());
		fileWriter.close();
		return;
		
	}
	
	public static void writeAdditionalSignalsTimeseriesWithTarget(
		final File file,
		final List<? extends IDtoFloatDataPoint> targetDataPoints,
		final String targetSignalId
	) throws FileNotFoundException, IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		
		final ArrayNode json = MAPPER.createArrayNode();
		final ObjectNode timeseriesByIdNode = MAPPER.createObjectNode();
		json.add(timeseriesByIdNode);
		
		timeseriesByIdNode.put("id", targetSignalId);
		
		final ObjectNode timeseriesNode = MAPPER.createObjectNode();
		timeseriesByIdNode.set("timeseries", timeseriesNode);
		
		timeseriesNode.put("_type", "Float");
		final ArrayNode dataPointsNode = timeseriesNode.putArray("dataPoints");
		for (IDtoFloatDataPoint dataPoint : targetDataPoints) {
			final ObjectNode dataPointNode = MAPPER.createObjectNode();
			dataPointNode.put("timestamp", dataPoint.getTimestamp());
			dataPointNode.put("value", dataPoint.getValue());
			dataPointsNode.add(dataPointNode);
		}
		
		final FileWriter fileWriter = new FileWriter(file);
		fileWriter.write(json.toPrettyString());
		fileWriter.close();
		
		return;
		
	}
	
	public static <T> void writeJson(final File file, final T obj) throws JsonGenerationException, JsonMappingException, IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		Optional.ofNullable(file.getParentFile()).ifPresent(f -> f.mkdirs());
		MAPPER.writeValue(file, obj);
	}
	
	public static Double computeR2(final List<Double> trueLabelValues, final List<Double> predictions) {
		final Double trueAvg = trueLabelValues.stream().mapToDouble(d -> d).sum() / trueLabelValues.size();
		final Double RSS = IntStream.range(0, predictions.size()).mapToDouble(i -> Math.pow(predictions.get(i) - trueLabelValues.get(i), 2)).sum();
		final Double TSS = IntStream.range(0, predictions.size()).mapToDouble(i -> Math.pow(trueLabelValues.get(i) - trueAvg, 2)).sum();
		return 1 - RSS / TSS;
	}
	
}
