# this python script performs the same calculation as example_sp.ipynb

import json
import argparse
import os

from aivis_engine_v2_sp_sdk_python import (
    SignalPredictionSetup,
    SignalPredictionData,
    SignalPredictionTraining,
    SignalPredictionInference,
    EngineLogger,
)

# ---------------------------------------------------------------------------------
# Data location

TRAINING_DATA_FILE = "train_sp.csv"
INFERENCE_DATA_FILE = "eval_sp.csv"
TARGET_SIGNAL_ID = "TARGET"


# Input and output folder specified via args
def from_args():
    parser = argparse.ArgumentParser(
        description="Example project for signal prediction"
    )
    # data is expected in input folder
    parser.add_argument(
        "--input", "-i", type=str, nargs=1, required=True, help="Path to input folder"
    )
    # output files will be written to output folder
    parser.add_argument(
        "--output", "-o", type=str, nargs=1, required=True, help="Path to output folder"
    )
    args = parser.parse_args()
    # create output folder if not existing
    if not os.path.exists(args.output[0]):
        os.mkdir(args.output[0])
    return {
        "input_folder": args.input[0],
        "output_folder": args.output[0],
    }


# ---------------------------------------------------------------------------------
# Logger and Threads


class Logger(EngineLogger):
    def log(self, level, thread, module, message):
        min_logging_level = 3
        if level <= min_logging_level:
            print("\t... %s" % message)


def setup_logger_and_threads():
    print("Initializing SignalPredictionSetup: logging and threads ...")
    SignalPredictionSetup.register_logger(Logger())
    max_threads = 0  # number of threads will be chosen automatically with 0
    SignalPredictionSetup.init_thread_count(max_threads)
    print("\t... done!")
    return


# ---------------------------------------------------------------------------------
# Reading and Writing


def read_json(filename):
    print("\t... reading file '%s' ..." % filename)
    with open(filename, "r") as file:
        return json.loads(file.read())


def write_json(filename, obj):
    print("\t... writing file '%s' ..." % filename)
    with open(filename, "w") as file:
        file.write(json.dumps(obj, indent=2))
    return


# convert [DtoFloatDataPoint] to json
def float_data_points_to_timeseries_json(list_object):
    return {
        "_type": "Float",
        "dataPoints": [
            {
                "timestamp": int(point_object.timestamp),
                "value": float(point_object.value),
            }
            for point_object in list_object
        ],
    }


# ---------------------------------------------------------------------------------
# MAIN


def main():
    folder_config = from_args()  # parsing args as folder_config
    setup_logger_and_threads()

    print("Loading training data ...")
    training_data = SignalPredictionData.create()

    files_filter = {"includeFiles": [TRAINING_DATA_FILE]}
    files_reader_config = json.dumps(
        {
            "folder": folder_config["input_folder"],
            "filesFilter": files_filter,
        }
    )

    training_data.read_files(files_reader_config)

    # print some signal to illustrate type DtoFloatDataPoint
    print("Example signal!")
    some_signal_id = "SIGNAL_7"
    print("\t... Signal id:", some_signal_id)
    some_signal_data_points = training_data.get_float_signal(some_signal_id)
    print("\t... Signal length:", len(some_signal_data_points))
    print("\t... Signal entry number 42:", some_signal_data_points[42])
    print(
        "\t... Signal entry number 42 timestamp (unix/millisec):",
        some_signal_data_points[42].timestamp,
    )
    print("\t... Signal entry number 42 value:", some_signal_data_points[42].value)

    # ---- TRAINING ----

    print("Creating training ...")
    # build training config - for more options see documentation
    training_config = json.dumps(
        {
            "target": {"signal": TARGET_SIGNAL_ID},
        }
    )

    # create training and train the model
    training = SignalPredictionTraining.create(training_data, training_config)
    print("\t... done!")

    print("Save training files ...")
    model = json.loads(training.get_model())
    write_json(os.path.join(folder_config["output_folder"], "model.json"), model)

    training_report = json.loads(training.get_report())  # training infos
    write_json(
        os.path.join(folder_config["output_folder"], "training-report.json"),
        training_report,
    )
    print("\t... done!")

    # ---- INFERENCE ----
    print("Loading inference data ...")
    inference_data = SignalPredictionData.create()

    files_filter = {"includeFiles": [INFERENCE_DATA_FILE]}
    data_filter = {
        "_type": "Inference",
        "dataSpecifications": [model["inferenceDataSpecification"]],
        "additionalSignals": [TARGET_SIGNAL_ID],
        "inferenceTimeRanges": [{}],
    }
    files_reader_config = json.dumps(
        {
            "folder": folder_config["input_folder"],
            "filesFilter": files_filter,
            "dataFilter": data_filter,
        }
    )

    inference_data.read_files(files_reader_config)

    print("Creating inference ...")
    # build inference config
    inference_config = json.dumps({"skipOnInsufficientData": True})
    # via model is also an option: SignalPredictionInference.create_by_model(model_json, inference_config)
    inference = SignalPredictionInference.create_by_training(training, inference_config)
    print("\t... done!")

    print("Calculating predictions ...")
    # choose inference timestamps - exactly where the target has a data point to get comparable target-prediction-pairs for kpis
    # but of course any timestamp where the inference data meets the model's inference data specification is possible
    target = inference_data.get_float_signal(TARGET_SIGNAL_ID)
    timestamps = list(map(lambda p: p.timestamp, target))

    # infer predictions
    predictions = inference.infer_float(inference_data, timestamps)

    print("Save inference files ...")
    write_json(
        os.path.join(folder_config["output_folder"], "predictions.json"),
        float_data_points_to_timeseries_json(predictions),
    )
    write_json(
        os.path.join(folder_config["output_folder"], "additional-signals.json"),
        [
            {
                "id": TARGET_SIGNAL_ID,
                "timeseries": float_data_points_to_timeseries_json(target),
            }
        ],
    )
    print("\t... done!")

    print("Destroy everything ...")
    # destroying can be done once the respective object is not needed anymore - do this if memory is an issue
    training_data.destroy()
    training.destroy()
    inference_data.destroy()
    inference.destroy()
    print("\t... done!")
    # aivis part is

    # ---- KPI: compute r² value of inference----

    # get list of predictions and timestamps where predictions exist
    # note that predictions might not exist at all timestamps where target exists due to missing predecessor in features
    pred_values = list(map(lambda p: p.value, predictions))
    pred_timestamps = list(map(lambda p: p.timestamp, predictions))

    # get target values at timestamps_pred
    true_target = list(filter(lambda p: p.timestamp in pred_timestamps, target))
    true_target_values = list(map(lambda p: p.value, true_target))

    true_avg = sum(true_target_values) / len(true_target_values)
    rss = sum(
        [(true - pred) ** 2 for (true, pred) in zip(true_target_values, pred_values)]
    )
    tss = sum([(true - true_avg) ** 2 for true in true_target_values])

    print("-----------------------------------------")
    print("R² value of inference: ", 1 - rss / tss)
    print("-----------------------------------------")

    print("Generating report ...")

    # ---- Toolbox HTML Report ----
    from aivis_engine_v2_toolbox.api import build_signal_prediction_report

    build_signal_prediction_report(
        {
            "title": "Aivis Engine V2 - Signal Prediction Example: Paper Quality",
            "predictions": {
                "file": os.path.join(folder_config["output_folder"], "predictions.json")
            },
            "target": {
                "file": os.path.join(
                    folder_config["output_folder"], "additional-signals.json"
                )
            },
            "trainingReportFile": os.path.join(
                folder_config["output_folder"], "training-report.json"
            ),
            "outputFile": os.path.join(
                folder_config["output_folder"],
                "aivis-engine-v2-signal-prediction-example.html",
            ),
        }
    )

    print("ALL DONE!")
    return


# run main function
main()
